import sys
import logging
from typing import Optional

from .thread import ThreadPool
from .process import ProcessPool

if sys.version_info >= (3, 7):
    from .ndprocess import NdProcessPool
else:
    NdProcessPool = None

try:
    from gevent.monkey import is_anything_patched

    if not is_anything_patched():
        raise ImportError("gevent did not patch anything")

    from .gevent import GreenletPool
except ImportError:
    # gevent is missing or didn't monkey patch the environment
    GreenletPool = None

    from .multiprocessing import MProcessPool
    from .ndmultiprocessing import NProcessPool

    try:
        from .billiard import BProcessPool
    except ImportError:
        BProcessPool = None
else:
    MProcessPool = None
    NProcessPool = None
    BProcessPool = None

from .scaling import ScalingPool

logger = logging.getLogger(__name__)

_POOLS = {
    "thread": ThreadPool,
    "process": ProcessPool,
    "ndprocess": NdProcessPool,
    "multiprocessing": MProcessPool,
    "ndmultiprocessing": NProcessPool,
    "billiard": BProcessPool,
    "gevent": GreenletPool,
    "scaling": ScalingPool,
}

_MESSAGES = {
    "ndprocess": "not supported in python < 3.6",
    "billiard": (
        "requires 'billiard'"
        if GreenletPool is None
        else "does not work with 'gevent' monkey patching"
    ),
    "gevent": "requires 'gevent' with monkey patching",
    "multiprocessing": "does not work with 'gevent' monkey patching",
    "ndmultiprocessing": "does not work with 'gevent' monkey patching",
}


def get_pool(pool_type: Optional[str] = None):
    if pool_type is None:
        if GreenletPool is None:
            pool_type = "process"
        else:
            pool_type = "gevent"
    pool = _POOLS[pool_type]
    if pool is None:
        raise ImportError(
            _MESSAGES.get(pool_type, f"pool type '{pool_type}' is unknown")
        )
    logger.info(f"pypushflow concurrency: {pool_type}")
    return pool



import time
import subprocess
from contextlib import contextmanager
from typing import Any, Optional, Tuple

import pytest
from concurrent.futures import ProcessPoolExecutor as CfPool
from multiprocessing.pool import Pool as MpPool
from multiprocessing import get_context as get_mp_context
from multiprocessing import get_all_start_methods

try:
    from billiard.pool import Pool as BPool
    from billiard import get_context as get_billiard_context
except ImportError:
    BPool = None
    BProcess = None

from ...concurrent import get_pool
from ...concurrent import exceptions
from ...concurrent.factory import _POOLS


CONTEXTS = [None] + get_all_start_methods()
POOLS = [s for s in _POOLS if s and s != "scaling"]
TIMEOUT = 10


@contextmanager
def pool_context(scaling, pool_type, **pool_options):
    cpool_type = pool_type
    if scaling:
        cpool_type = "scaling"
        try:
            get_pool(pool_type)
        except ImportError as e:
            pytest.skip(str(e))
    try:
        pool_cls = get_pool(cpool_type)
    except ImportError as e:
        pytest.skip(str(e))
    pool_options.setdefault("wait_on_exit_timeout", TIMEOUT)
    with pool_cls(pool_type=pool_type, **pool_options) as pool:
        yield pool


def assert_callback(pool, name):
    libname, fn, result = SUCCESS[name]
    if fn is None:
        pytest.skip(f"requires {libname}")
    value, exception = _submit_with_cb(pool, fn)
    if exception is not None:
        raise exception
    assert value == result, str(value)


def assert_error_callback(pool, name):
    libname, fn, result = FAILURE[name]
    if fn is None:
        pytest.skip(f"requires {libname}")
    value, exception = _submit_with_cb(pool, fn)
    assert value is None
    assert str(exception) == str(result), str(value)
    assert isinstance(exception, type(result))

    tb = exceptions.serialize_exception(exception)["traceBack"]
    msg = "\n\n" + "".join(tb)
    assert any("in _fn_failure" in s for s in tb), msg
    assert any('raise RuntimeError("expected failure")' in s for s in tb), msg


def _submit_with_cb(pool, fn) -> Tuple[Any, Optional[Exception]]:
    value = None
    exception = None
    ev = Event()

    def cb(result):
        nonlocal value
        value = result
        ev.set()

    def ecb(result):
        nonlocal exception
        exception = result
        ev.set()

    pool.apply_async(fn, callback=cb, error_callback=ecb)
    assert ev.wait(timeout=TIMEOUT)

    return value, exception


def _fn_success() -> int:
    return 10


def _fn_failure():
    raise RuntimeError("expected failure")


def _fn_subprocess_success() -> int:
    return int(subprocess.check_output(["echo", "10"]).decode().strip())


def _fn_mppool_success() -> int:
    with MpPool() as pool:
        return pool.apply(_fn_success)


def _fn_mpprocess_success() -> int:
    ctx = get_mp_context()
    result = ctx.Manager().Value("i", 0)
    p = ctx.Process(target=_fn_mpprocess_success_helper, args=(result,))
    p.start()
    p.join()
    return result.value


def _fn_mpprocess_success_helper(result) -> None:
    result.value = 10


def _fn_cfpool_success() -> int:
    with CfPool() as pool:
        return pool.submit(_fn_success).result()


class Event:
    def __init__(self) -> None:
        self._is_set = False

    def set(self):
        self._is_set = True

    def wait(self, timeout=None):
        t0 = time.time()
        while not self._is_set:
            if timeout is not None:
                if (time.time() - t0) > timeout:
                    return False
            time.sleep(0.2)
        return True

    def reset(self):
        self._is_set = False


if BPool is None:
    _fn_bpool_success = None
    _fn_bprocess_success = None
else:

    def _fn_bpool_success() -> int:
        with BPool() as pool:
            return pool.apply(_fn_success)

    def _fn_bprocess_success() -> int:
        ctx = get_billiard_context()
        result = ctx.Manager().Value("i", 0)
        p = get_billiard_context().Process(
            target=_fn_bprocess_success_helper, args=(result,)
        )
        p.start()
        p.join()
        return result.value

    def _fn_bprocess_success_helper(result) -> None:
        result.value = 10


SUCCESS = {
    "simple": ("builtins", _fn_success, 10),
    "subprocess": ("subprocess", _fn_subprocess_success, 10),
    "cfpool": ("concurrent.futures", _fn_cfpool_success, 10),
    "mppool": ("multiprocessing", _fn_mppool_success, 10),
    "mpprocess": ("multiprocessing", _fn_mpprocess_success, 10),
    "bpool": ("billiard", _fn_bpool_success, 10),
    "bprocess": ("billiard", _fn_bprocess_success, 10),
}

FAILURE = {"simple": ("builtins", _fn_failure, RuntimeError("expected failure"))}



from numbers import Number
import logging
from billiard.pool import Pool as _Pool
from billiard.pool import ApplyResult as Future
from billiard import get_context
from typing import Callable, Optional
from . import base
from .interrupt import process as interrupt


logger = logging.getLogger(__name__)


class RemoteTraceback(Exception):
    def __init__(self, tb):
        self.tb = tb

    def __str__(self):
        return self.tb


class BProcessPool(base.BasePool):
    """Pool of non-daemonic processes (but they CAN have sub-processes)."""

    def __init__(
        self, context: str = None, max_workers: Optional[int] = None, **kw
    ) -> None:
        kwargs = dict()
        if isinstance(context, str) or None:
            context = get_context(context)
        logger.info(f"pypushflow process pool context: '{type(context).__name__}'")
        kwargs["context"] = context
        if max_workers is not None:
            kwargs["processes"] = max_workers
        self._pool = _Pool(initializer=interrupt.worker_initializer, **kwargs)
        super().__init__(**kw)

    def __enter__(self):
        self._pool.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self._pool.__exit__(exc_type, exc_val, exc_tb)

    def shutdown(
        self,
        block: bool = False,
        timeout: Optional[Number] = None,
        interrupt: bool = False,
    ) -> None:
        super().shutdown(block=block, timeout=timeout, interrupt=interrupt)
        self._pool.terminate()

    def close(self):
        self._pool.close()

    def join(self, timeout: Optional[Number] = None) -> bool:
        self._pool.join()
        return True

    def interrupt(self) -> None:
        for p in list(self._pool._pool):
            interrupt.interrupt_task(p.pid)

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ) -> Future:
        if kwargs is None:
            kwargs = dict()

        if args is None:
            args = (fn,)
        else:
            args = (fn,) + args

        if error_callback is not None:
            org_error_callback = error_callback

            def error_callback(info):
                info.exception.__cause__ = RemoteTraceback(info.traceback)
                return org_error_callback(info.exception)

        return self._pool.apply_async(
            interrupt.task_main,
            args=args,
            kwds=kwargs,
            callback=callback,
            error_callback=error_callback,
        )



import unittest

from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ErrorHandler import ErrorHandler
from pypushflow.ForkActor import ForkActor
from pypushflow.JoinActor import JoinActor
from pypushflow.RouterActor import RouterActor
from pypushflow.ThreadCounter import ThreadCounter


class TestPythonActor(unittest.TestCase):
    def setUp(self):
        self.thread_counter = ThreadCounter()

    def test_PythonActor(self):
        script = "pypushflow.tests.tasks.pythonActorTest.py"
        name = "Python Actor Test"
        actor = PythonActor(
            script=script, name=name, thread_counter=self.thread_counter
        )
        stopActor = StopActor(thread_counter=self.thread_counter)
        inData = {"name": "Ragnar"}
        actor.connect(stopActor)
        actor.trigger(inData)
        stopActor.join(timeout=10)
        outData = stopActor.outData
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Ragnar!")

    def test_PythonActorWithPositionalArguments(self):
        script = "pypushflow.tests.tasks.pythonActorTest.py"
        name = "Python Actor Test"
        actor = PythonActor(
            script=script, name=name, thread_counter=self.thread_counter
        )
        stopActor = StopActor(thread_counter=self.thread_counter)
        inData = {0: "Ragnar"}
        actor.connect(stopActor)
        actor.trigger(inData)
        stopActor.join(timeout=10)
        outData = stopActor.outData
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Ragnar!")

    def test_ErrorHandler(self):
        script = "pypushflow.tests.tasks.pythonErrorHandlerTest.py"
        name = "Python Error Handler Test"
        actor = PythonActor(
            script=script, name=name, thread_counter=self.thread_counter
        )
        stopActor = StopActor(thread_counter=self.thread_counter)
        errorHandler = ErrorHandler(
            name="Error handler", thread_counter=self.thread_counter
        )
        inData = {"name": "Ragnar"}
        actor.connect(stopActor)
        actor.connectOnError(errorHandler)
        errorHandler.connect(stopActor)
        actor.trigger(inData)
        stopActor.join(timeout=5)
        outData = stopActor.outData
        self.assertIsNotNone(outData)
        self.assertTrue("WorkflowException" in outData)

    def test_ForkAndJoinActors(self):
        start = StartActor(thread_counter=self.thread_counter)
        stop = StopActor(thread_counter=self.thread_counter)
        fork = ForkActor(thread_counter=self.thread_counter)
        joinActor = JoinActor(thread_counter=self.thread_counter)
        pythonActor1 = PythonActor(
            script="pypushflow.tests.tasks.pythonActor1.py",
            thread_counter=self.thread_counter,
        )
        pythonActor2 = PythonActor(
            script="pypushflow.tests.tasks.pythonActor2.py",
            thread_counter=self.thread_counter,
        )
        # Connections
        start.connect(fork)
        fork.connect(pythonActor1)
        fork.connect(pythonActor2)
        pythonActor1.connect(joinActor)
        joinActor.increaseNumberOfThreads()
        pythonActor2.connect(joinActor)
        joinActor.increaseNumberOfThreads()
        joinActor.connect(stop)
        # Run
        inData = {"a": 1}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)

    def test_RouterActor(self):
        start = StartActor(thread_counter=self.thread_counter)
        stop = StopActor(thread_counter=self.thread_counter)
        router = RouterActor(
            parent=None,
            itemName="a",
            listPort=["other", "null"],
            thread_counter=self.thread_counter,
        )
        pythonActor1 = PythonActor(
            script="pypushflow.tests.tasks.pythonActor1.py",
            thread_counter=self.thread_counter,
        )
        pythonActor2 = PythonActor(
            script="pypushflow.tests.tasks.pythonActor2.py",
            thread_counter=self.thread_counter,
        )
        # Connections
        start.connect(router)
        router.connect(pythonActor1, "other")
        router.connect(pythonActor2, "null")
        pythonActor1.connect(stop)
        pythonActor2.connect(stop)
        # Run
        inData = {"a": 1}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)
        self.assertTrue(outData["actor1"])
        # Run 2
        inData = {"a": None}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)
        self.assertTrue(outData["actor2"])



from contextlib import contextmanager
from multiprocessing import get_context

try:
    from multiprocessing import get_context as bget_context
except ImportError:
    bget_context = None
import pytest

from ..conftest import gevent_patched


@contextmanager
def filter_callback(pool_type, task_name, context):
    assert pool_type
    assert task_name
    exception = None
    match = None
    context = _context_name(pool_type, context)
    gevent_incompatible_tasks = ("mppool", "bpool", "mpprocess", "bprocess")
    if pool_type in ("billiard", "multiprocessing"):
        if gevent_patched():
            pytest.skip("does not work with 'gevent' monkey patching")
        if task_name in ["mppool", "mpprocess", "cfpool"]:
            exception = AssertionError
            match = "daemonic processes are not allowed to have children"
        elif task_name in ["bpool"]:
            pytest.skip("hangs sometimes")
    elif pool_type == "ndmultiprocessing":
        if gevent_patched():
            pytest.skip("does not work with 'gevent' monkey patching")
        if task_name in ["bpool"]:
            pytest.skip("hangs sometimes")
    elif pool_type in ("process", "ndprocess"):
        if gevent_patched():
            if task_name in gevent_incompatible_tasks:
                pytest.skip("task not compatible with gevent")
            if context == "spawn":
                pytest.skip("spawn hangs with gevent")
        else:
            if task_name in ["bpool"]:
                pytest.skip("hangs sometimes")
    elif pool_type == "thread":
        if gevent_patched():
            if task_name in gevent_incompatible_tasks:
                pytest.skip("task not compatible with gevent")
        else:
            if task_name in ["bpool"]:
                pytest.skip("hangs sometimes")
    elif pool_type == "gevent":
        if task_name in gevent_incompatible_tasks:
            pytest.skip("task not compatible with gevent")

    if exception is None:
        yield
    else:
        with pytest.raises(exception, match=match):
            yield


@contextmanager
def filter_error_callback(pool_type, task_name, context):
    assert pool_type
    assert task_name
    context = _context_name(pool_type, context)
    if pool_type in ("process", "ndprocess"):
        if gevent_patched():
            if context == "spawn":
                pytest.skip("spawn hangs with gevent")
    yield


def _context_name(pool_type, context):
    if context is not None:
        return context
    if pool_type in "billiard":
        return bget_context()._name
    else:
        return get_context()._name



from numbers import Number
from typing import Callable, Optional


class BasePool:
    def __init__(
        self,
        wait_on_exit: bool = True,
        wait_on_exit_timeout: Optional[Number] = None,
        interrupt_on_exit: bool = False,
        **_
    ) -> None:
        self._wait_on_exit = wait_on_exit
        self._wait_on_exit_timeout = wait_on_exit_timeout
        self._interrupt_on_exit = interrupt_on_exit

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(
            block=self._wait_on_exit,
            timeout=self._wait_on_exit_timeout,
            interrupt=self._interrupt_on_exit,
        )

    def close(self):
        """Prevents any more tasks from being submitted to the pool."""
        raise NotImplementedError

    def join(self, timeout: Optional[Number] = None):
        """Wait for the workers to exit."""
        raise NotImplementedError

    def shutdown(
        self,
        block: bool = False,
        timeout: Optional[Number] = None,
        interrupt: bool = False,
    ) -> None:
        """Cleanup all resources. Waits for tasks to finish unless `interrupt=True`."""
        self.close()
        if interrupt:
            self.interrupt()
        if block:
            self.join(timeout=timeout)

    def interrupt(self) -> None:
        """Stop all running tasks"""
        raise NotImplementedError

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ):
        raise NotImplementedError



import os
import sys
import time
import signal
import multiprocessing as mp
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor

import pytest
from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.ErrorHandler import ErrorHandler
from pypushflow.stop_workflows import DEFAULT_STOP_SIGNALS


@pytest.mark.parametrize("forced_interruption", [True, False])
def test_stop_workflow(forced_interruption, workflow_cleanup):
    if os.name == "nt" and forced_interruption:
        pytest.skip("not supported on windows")
    testWorkflow1 = WorkflowSleep("Test workflow Sleep")

    def run_normal(executor):
        # Workflow with 3 nodes that sleep 0 seconds each.
        inData = {"sleep_time": 0, "counter": 0}
        future = executor.submit(
            testWorkflow1.run,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
        )
        result = future.result()
        assert "WorkflowException" not in result
        assert result["counter"] == 3

    def run_stopped(executor):
        # Workflow with 3 nodes that sleep 2 seconds each.
        inData = {"sleep_time": 2, "counter": 0}
        future = executor.submit(
            testWorkflow1.run,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
        )

        # Stop half-way the execution of the second node.
        time.sleep(3)
        print("STOP ...")
        testWorkflow1.stop(
            reason="workflow stopped by user", forced_interruption=forced_interruption
        )

        result = future.result()
        assert result["counter"] < 3
        assert "WorkflowException" in result

    with ThreadPoolExecutor(max_workers=1) as executor:
        run_normal(executor)
        run_stopped(executor)
        run_stopped(executor)
        run_normal(executor)


@pytest.mark.parametrize("forced_interruption", [True, False])
def test_stop_signal_workflow(forced_interruption, skip_when_gevent, skip_on_windows):
    def run_normal(executor):
        # Workflow with 3 nodes that sleep 0 seconds each.
        inData = {"sleep_time": 0, "counter": 0}
        future = executor.submit(
            _run_workflow,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
            forced_interruption=forced_interruption,
        )
        result = future.result()
        assert "WorkflowException" not in result
        assert result["counter"] == 3

    def run_stopped(executor, started_event):
        # Workflow with 3 nodes that sleep 2 seconds each.
        inData = {"sleep_time": 2, "counter": 0}
        started_event.clear()
        future = executor.submit(
            _run_workflow,
            inData,
            timeout=15,
            scaling_workers=False,
            max_workers=-1,
            started_event=started_event,
            forced_interruption=forced_interruption,
        )

        # Stop signal half-way the execution of the second node.
        assert started_event.wait(timeout=10)
        time.sleep(3)
        for pid in list(executor._processes):
            # TODO: doesn't work on windows (it kills the subprocess)
            print(f"KILL pid={pid} ...")
            os.kill(pid, DEFAULT_STOP_SIGNALS[0])

        result = future.result()
        assert result["counter"] < 3
        assert "WorkflowException" in result

    with mp.Manager() as manager:
        started_event = manager.Event()

        if sys.version_info < (3, 7):
            pool = ProcessPoolExecutor(max_workers=1)
            pool.submit(_process_initializer).result()
        else:
            pool = ProcessPoolExecutor(max_workers=1, initializer=_process_initializer)

        with pool as executor:
            print("\n--run_normal--")
            run_normal(executor)
            print("\n--run_stopped--")
            run_stopped(executor, started_event)
            print("\n--run_stopped--")
            run_stopped(executor, started_event)
            print("\n--run_normal--")
            run_normal(executor)


def _process_initializer():
    try:
        signal.signal(DEFAULT_STOP_SIGNALS[0], signal.SIG_IGN)
    except (OSError, AttributeError, ValueError, RuntimeError):
        pass


def _run_workflow(
    *args,
    started_event=None,
    forced_interruption: bool = False,
    **kwargs,
):
    print(f"START WORKFLOW pid={os.getpid()}")
    testWorkflow1 = WorkflowSleep(
        "Test workflow Sleep",
        stop_on_signals=True,
        forced_interruption=forced_interruption,
    )
    if started_event is not None:
        started_event.set()
    return testWorkflow1.run(*args, **kwargs)


class WorkflowSleep(Workflow):
    def __init__(self, name, **kw):
        super().__init__(name, **kw)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.errorActor = ErrorHandler(parent=self, thread_counter=ctr)
        self.pythonActor1 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.pythonActor2 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.pythonActor3 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorInterrupt.py",
            name="Python Actor Sleep",
            thread_counter=ctr,
        )
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.pythonActor1)
        self.pythonActor1.connect(self.pythonActor2)
        self.pythonActor2.connect(self.pythonActor3)
        self.pythonActor3.connect(self.stopActor)
        self.errorActor.connect(self.stopActor)
        self.pythonActor1.connectOnError(self.errorActor)
        self.pythonActor2.connectOnError(self.errorActor)
        self.pythonActor3.connectOnError(self.errorActor)



"""Persistent recording of a workflow executions."""

import os
import warnings
from typing import Optional, Callable

from .interface import WorkflowDbClient
from .pymongo import PyMongoWorkflowDbClient  # noqa F401
from .besdb import BesWorkflowDbClient  # noqa F401
from .mongita import MemoryWorkflowDbClient  # noqa F401
from .dummy import DummyWorkflowDbClient  # noqa F401

DEFAULT_DB_TYPE = "dummy"


def init_db_client(*args, db_type: Optional[str] = None, **kwargs) -> WorkflowDbClient:
    """Initializes a database client based on the specified `db_type`.

    :param db_type: The type of database client to initialize.
                    If not specified, defaults to `DEFAULT_DB_TYPE`.
                    Supported values include:

                    - **"besdb"**: Requires the following additional parameters:
                      - **url** (str): URL for connecting to the BES database.
                      - **initiator** (str): Initiator from which the request originates.
                      - **host** (str): Hostname from which the request originates.
                      - **port** (int): Port number from which the request originates.
                      - **request_id** (str): Unique identifier for the request.

                    - **"pymongo"**: Requires the following additional parameters:
                      - **url** (str): Connection URL for the MongoDB instance.
                      - **database** (str): Name of the database to access.
                      - **collection** (str): Name of the collection within the database.

                    - **"memory"**: An in-memory database type. No additional parameters required.

                    - **"dummy"**: A placeholder database type for testing or development purposes. No additional parameters required.

    :param args: see `db_type`.
    :param kwargs: see `db_type`.

    :return: An instance of `WorkflowDbClient` specific to the specified `db_type`.
    """

    if db_type is None:
        url = os.environ.get("PYPUSHFLOW_MONGOURL", None)
        if url:
            warnings.warn(
                "Using BESDB environment variables is deprecated", DeprecationWarning
            )
            db_type = "besdb"
            kwargs = {
                "url": url,
                "initiator": os.environ.get("PYPUSHFLOW_INITIATOR", "unknown"),
                "host": os.environ.get("PYPUSHFLOW_HOST", "unknown"),
                "port": os.environ.get("PYPUSHFLOW_PORT", "unknown"),
                "request_id": os.environ.get("PYPUSHFLOW_OBJECTID", "unknown"),
            }
        else:
            db_type = DEFAULT_DB_TYPE
    db_client_class = WorkflowDbClient.get_dbclient_class(db_type)
    return db_client_class(*args, **kwargs)


def register_actorinfo_filter(method: Callable):
    WorkflowDbClient.register_actorinfo_filter(method)



import os
import time
from numbers import Number
import pytest
from . import utils
from ...concurrent.interrupt import StopPypushflowTask


def create_file(sleep_time: Number, filename: str) -> str:
    time.sleep(sleep_time)
    with open(filename, "w"):
        pass
    return filename


@pytest.mark.parametrize("scaling", [True, False])
@pytest.mark.parametrize("max_workers", [None, 1])
@pytest.mark.parametrize("pool_type", utils.POOLS)
def test_interrupt(scaling, max_workers, pool_type, tmpdir):
    if pool_type == "thread":
        pytest.skip("threads cannot be interrupted")
    if os.name == "nt" and pool_type in (
        "process",
        "ndprocess",
        "multiprocessing",
        "ndmultiprocessing",
        "billiard",
    ):
        pytest.skip("not supported on windows")

    callback_event = utils.Event()
    exception = None
    result = None

    def reset():
        nonlocal result
        nonlocal exception
        result = None
        exception = None
        callback_event.reset()

    def callback(r):
        nonlocal result
        result = r
        callback_event.set()

    def error_callback(e):
        nonlocal exception
        exception = e
        callback_event.set()

    def run_normal(filename):
        nonlocal result
        reset()
        pool.apply_async(
            create_file,
            args=(0, str(filename)),
            callback=callback,
            error_callback=error_callback,
        )
        callback_event.wait(10)
        assert result == str(filename)
        assert filename.exists()

    def run_interrupt(filename):
        nonlocal exception
        reset()
        pool.apply_async(
            create_file,
            args=(5, str(filename)),
            callback=callback,
            error_callback=error_callback,
        )
        time.sleep(2)
        pool.interrupt()
        callback_event.wait(10)
        assert isinstance(exception, StopPypushflowTask), str(type(exception))
        assert not filename.exists()

    with utils.pool_context(scaling, pool_type, max_workers=max_workers) as pool:
        run_normal(tmpdir / "test1.txt")
        run_interrupt(tmpdir / "test2.txt")
        run_interrupt(tmpdir / "test3.txt")
        run_normal(tmpdir / "test4.txt")



from pypushflow.AbstractActor import AbstractActor


class RequestStatus(AbstractActor):
    def __init__(self, parent=None, name="Request status", status=None, **kw):
        super().__init__(parent=parent, name=name, **kw)
        self.status = status

    def trigger(self, inData):
        if self.status is not None and self.parent is not None:
            self.parent.setStatus(self.status)
        super().trigger(inData=inData)



import logging
import hashlib

from .pymongo import PyMongoWorkflowDbClient

try:
    from bson.objectid import ObjectId
except Exception:
    ObjectId = None


_logger = logging.getLogger(__name__)


class BesWorkflowDbClient(PyMongoWorkflowDbClient, register_name="besdb"):
    """Client of the BES Mongo database for storing workflow executions."""

    def __init__(self, url: str, initiator: str, host: str, port: int, request_id: str):
        self._request_id = _create_objectid(str(request_id))

        self._initial_workflow_info = {
            "_id": self._request_id,
            "initiator": str(initiator),
            "host": str(host),
            "port": str(port),
            "Request ID": str(request_id),
        }

        super().__init__(url, "bes", "bes")

    def _generateInitialWorkflowInfo(self) -> dict:
        return dict(self._initial_workflow_info)

    def generateWorkflowId(self) -> ObjectId:
        return self._request_id


def _create_objectid(request_id: str) -> ObjectId:
    try:
        return ObjectId(request_id)
    except Exception:
        pass

    hash_bytes = hashlib.blake2s(request_id.encode(), digest_size=12).digest()
    object_id = ObjectId(hash_bytes)

    _logger.warning(
        "'%s' is not a valid BSON Object ID. Use its hash '%s' instead.",
        request_id,
        object_id,
    )

    return object_id



import os
import pytest
import psutil
import logging


logger = logging.getLogger(__name__)


@pytest.fixture()
def workflow_cleanup():
    current_process = psutil.Process()
    subprocesses = {child.pid for child in current_process.children(recursive=True)}
    fail_on_subprocesses = True
    yield
    has_new_subprocesses = False
    for child in current_process.children(recursive=True):
        if child.pid not in subprocesses:
            has_new_subprocesses = True
            logger.error("dangling child %s", child)
            try:
                child.terminate()
            except psutil.NoSuchProcess:
                pass
    if has_new_subprocesses and fail_on_subprocesses:
        raise RuntimeError("not all subprocesses have finished")


def gevent_patched() -> bool:
    try:
        from gevent.monkey import is_anything_patched
    except ImportError:
        return False

    return is_anything_patched()


@pytest.fixture(scope="session")
def skip_when_gevent():
    if gevent_patched():
        pytest.skip("does not work with 'gevent' monkey patching")


@pytest.fixture(scope="session")
def skip_on_windows():
    if os.name == "nt":
        pytest.skip("does not work on windows")



import os
import pprint
import datetime
import importlib

from .AbstractActor import AbstractActor
from .concurrent import exceptions


class PythonActor(AbstractActor):
    def __init__(
        self, parent=None, name="Python Actor", errorHandler=None, script=None, **kw
    ):
        super().__init__(parent=parent, name=name, **kw)
        self.parentErrorHandler = errorHandler
        self.listErrorHandler = []
        self.script = script
        self.inData = None

    def connectOnError(self, actor):
        self.logger.debug("connect to error handler '%s'", actor.name)
        self.listErrorHandler.append(actor)

    def trigger(self, inData: dict):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        self.setStarted()
        self.inData = dict(inData)
        self.uploadInDataToMongo(actorData={"inData": inData}, script=self.script)

        try:
            module = importlib.import_module(os.path.splitext(self.script)[0])
        except Exception as e:
            self.logger.error("Error when trying to import script '%s'", self.script)
            self.errorHandler(e)
            return

        with self._postpone_end_thread(self.resultHandler, self.errorHandler) as (
            resultHandler,
            errorHandler,
        ):
            target = module.run
            self.logger.debug(
                "%s execution of '%s.%s'",
                "synchronous" if self.pool is None else "asynchronous",
                target.__module__,
                target.__name__,
            )

            args = {k: v for k, v in self.inData.items() if not isinstance(k, str)}
            kwargs = {k: v for k, v in self.inData.items() if isinstance(k, str)}
            if args:
                args = tuple(args.get(i) for i in range(max(args) + 1))
            else:
                args = tuple()

            if self.parent is not None:
                stop_exception = self.parent.stop_exception
                if stop_exception:
                    self.logger.error(str(stop_exception))
                    errorHandler(stop_exception)
                    return

            if self.pool is None:
                try:
                    result = target(*args, **kwargs)
                except BaseException as e:
                    errorHandler(e)
                else:
                    resultHandler(result)
            else:
                try:
                    self.pool.apply_async(
                        target,
                        args=args,
                        kwargs=kwargs,
                        callback=resultHandler,
                        error_callback=errorHandler,
                    )
                except RuntimeError:
                    if self.parent is not None:
                        stop_exception = self.parent.stop_exception
                        if stop_exception:
                            self.logger.error(str(stop_exception))
                            errorHandler(stop_exception)
                            return
                    raise

    @property
    def pool(self):
        if self.parent is not None:
            return self.parent.pool

    @property
    def pool_resources(self):
        return 1

    def resultHandler(self, result: dict):
        """Async callback in case of success"""
        try:
            # Handle the result
            self._finishedSuccess(result)

            # Trigger actors
            downstreamData = self.compileDownstreamData(result)
            self._triggerDownStreamActors(downstreamData)
        except Exception as e:
            self.errorHandler(e)

    def compileDownstreamData(self, result: dict) -> dict:
        return {**self.inData, **result}

    def errorHandler(self, exception: Exception):
        """Async callback in case of exception"""
        try:
            # Handle the result
            workflow_exception = exceptions.serialize_exception(exception)
            self._logException(exception)
            self._finishedFailure(workflow_exception)

            # Trigger actors
            result = {
                "WorkflowException": workflow_exception,
                "WorkflowExceptionInstance": exception,
            }
            downstreamData = self.compileDownstreamErrorData(result)
            self._triggerErrorHandlers(downstreamData)
        except Exception:
            self.logger.exception("In errorHandler for '%s'", self.name)

    def compileDownstreamErrorData(self, result: dict) -> dict:
        return {**self.inData, **result}

    def _triggerDownStreamActors(self, downstreamData: dict):
        for downStreamActor in self.listDownStreamActor:
            downStreamActor.trigger(downstreamData)

    def _triggerErrorHandlers(self, downstreamData: dict):
        for errorHandler in self.listErrorHandler:
            errorHandler.trigger(downstreamData)
        if self.parentErrorHandler is not None:
            self.parentErrorHandler.triggerOnError(inData=downstreamData)

    def _finishedSuccess(self, result: dict):
        self.setFinished()
        self.uploadOutDataToMongo(
            actorData={
                "stopTime": datetime.datetime.now(),
                "status": "finished",
                "outData": result,
            }
        )
        if "workflowLogFile" in result:
            self.setMongoAttribute("logFile", result["workflowLogFile"])
        if "workflowDebugLogFile" in result:
            self.setMongoAttribute("debugLogFile", result["workflowDebugLogFile"])

    def _finishedFailure(self, result: dict):
        self.setFinished()
        self.uploadOutDataToMongo(
            actorData={
                "stopTime": datetime.datetime.now(),
                "status": "error",
                "outData": result,
            }
        )

    def _logException(self, exception: Exception) -> None:
        if exception.__traceback__ is None:
            logfunc = self.logger.error
        else:
            logfunc = self.logger.exception
        logfunc(
            "Error in python actor '%s'!\n Not running down stream actors %s\n Exception:%s",
            self.name,
            [actor.name for actor in self.listDownStreamActor],
            exception,
        )



import sys
import numpy
import pytest
import mongomock

try:
    from bson.objectid import ObjectId
except Exception:
    ObjectId = None

from pypushflow.persistence import init_db_client
from pypushflow.persistence import WorkflowDbClient
from pypushflow.persistence import DummyWorkflowDbClient
from pypushflow.persistence import PyMongoWorkflowDbClient


@pytest.fixture()
def dummy_client():
    client = init_db_client(db_type="dummy")
    client.connect()
    return client


@pytest.fixture()
def memory_client():
    if sys.version_info < (3, 7):
        pytest.skip("mongita does not support python<3.7")
    client = init_db_client(db_type="memory")
    client.connect()
    yield client
    client.disconnect()


@pytest.fixture()
def mock_mongo():
    with mongomock.patch(on_new="create"):
        yield


@pytest.fixture()
def pymongo_client(mock_mongo):
    client = init_db_client(
        db_type="pymongo",
        url="mongodb://user:pwd@localhost:27017/mydatabase",
        database="mydatabase",
        collection="mycollection",
    )
    client.connect()
    yield client
    client.disconnect()


@pytest.fixture()
def besdb_client(mock_mongo):
    client = init_db_client(
        db_type="besdb",
        url="mongodb://user:pwd@localhost:27017/bes",
        initiator="PyPushflowUnitTests",
        host="unknown",
        port=-1,
        request_id=str(ObjectId()) if ObjectId else "unknown",
    )
    client.connect()
    yield client
    client.disconnect()


def test_workflow_dummy(dummy_client) -> None:
    _test_workflow_persistency(dummy_client)


def test_workflow_memory(memory_client) -> None:
    _test_workflow_persistency(memory_client)


def test_workflow_pymongo(pymongo_client) -> None:
    _test_workflow_persistency(pymongo_client)


def test_workflow_bes(besdb_client) -> None:
    _test_workflow_persistency(besdb_client)


def test_actor_dummy(dummy_client) -> None:
    _test_actor_persistency(dummy_client)


def test_actor_memory(memory_client) -> None:
    _test_actor_persistency(memory_client)


def test_actor_pymongo(pymongo_client) -> None:
    _test_actor_persistency(pymongo_client)


def test_actor_bes(besdb_client) -> None:
    _test_actor_persistency(besdb_client)


def _test_workflow_persistency(client: WorkflowDbClient) -> None:
    is_dummy = isinstance(client, DummyWorkflowDbClient)
    is_pymongo = isinstance(client, PyMongoWorkflowDbClient)

    name = "test_startWorkflow"
    client.startWorkflow(name)
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "started"

    client.updateWorkflowInfo({"status": "error"})
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "error"

    client.updateWorkflowInfo({"data": {"a": 1}})
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    else:
        assert info["data"], {"a": 1}

    client.updateWorkflowInfo({"data": {"a": 2}})
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    else:
        assert info["data"], {"a": 2}

    bad_data = {"exception": RuntimeError("error"), "numpy_array": numpy.arange(3)}
    client.updateWorkflowInfo(bad_data)
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    elif is_pymongo:
        for k, v in bad_data.items():
            assert info[k] == repr(v)
    else:
        for k, v in bad_data.items():
            assert type(info[k]) is type(v)
            assert repr(info[k]) == repr(v)

    client.endWorkflow()
    info = client.getWorkflowInfo()
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "error"


def _test_actor_persistency(client: WorkflowDbClient) -> None:
    is_dummy = isinstance(client, DummyWorkflowDbClient)
    is_pymongo = isinstance(client, PyMongoWorkflowDbClient)

    name = "test_startWorkflow"
    client.startWorkflow(name=name)

    actorName1 = "TestActor1"
    actorId1 = client.startActor(actorName1)
    info = client.getActorInfo(actorId1)
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "started"

    client.updateActorInfo(actorId1, {"status": "error"})
    info = client.getActorInfo(actorId1)
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "error"

    client.endActor(actorId1)
    info = client.getActorInfo(actorId1)
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "error"

    actorName2 = "TestActor2"
    actorId2 = client.startActor(name=actorName2)
    info = client.getActorInfo(actorId2)
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "started"

    client.updateActorInfo(actorId2, {"data": {"a": 1}})
    info = client.getActorInfo(actorId2)
    if is_dummy:
        assert info is None
    else:
        assert info["data"], {"a": 1}

    client.updateActorInfo(actorId2, {"data": {"a": 2}})
    info = client.getActorInfo(actorId2)
    if is_dummy:
        assert info is None
    else:
        assert info["data"], {"a": 1}

    client.endActor(actorId2)
    info = client.getActorInfo(actorId2)
    if is_dummy:
        assert info is None
    else:
        assert info["status"], "finished"

    actorName3 = "TestActor3"
    bad_data = {"exception1": RuntimeError("error1"), "numpy_array1": numpy.arange(3)}
    actorId3 = client.startActor(name=actorName3, info=bad_data)
    info = client.getActorInfo(actorId3)
    if is_dummy:
        assert info is None
    elif is_pymongo:
        # The type is unexpectidly preserved (might be because of mongomock)
        for k, v in bad_data.items():
            assert type(info[k]) is type(v)
            assert repr(info[k]) == repr(v)
    else:
        for k, v in bad_data.items():
            assert type(info[k]) is type(v)
            assert repr(info[k]) == repr(v)

    bad_data = {"exception2": RuntimeError("error2"), "numpy_array2": numpy.arange(4)}
    client.updateActorInfo(actorId3, bad_data)
    info = client.getActorInfo(actorId3)
    if is_dummy:
        assert info is None
    elif is_pymongo:
        for k, v in bad_data.items():
            assert info[k] == repr(v)
    else:
        for k, v in bad_data.items():
            assert type(info[k]) is type(v)
            assert repr(info[k]) == repr(v)



from typing import Mapping, Any

try:
    import bson
    import pymongo
    from bson.objectid import ObjectId
except Exception:
    bson = None
    pymongo = None
    ObjectId = None
from .mongo import MongoWorkflowDbClient


class PyMongoWorkflowDbClient(MongoWorkflowDbClient, register_name="pymongo"):
    """Client of an external Mongo database for storing workflow executions."""

    def __init__(self, url: str, database: str, collection: str):
        super().__init__()
        self._url = url
        self._database = database
        self._collection = collection

    def connect(self):
        if pymongo is None:
            return
        client = pymongo.MongoClient(self._url, serverSelectionTimeoutMS=10000)
        self._client = client
        self._collection = client[self._database][self._collection]

    def disconnect(self):
        self._collection = None
        if self._client is not None:
            self._client.close()
            self._client = None

    def generateWorkflowId(self) -> ObjectId:
        return ObjectId()

    def generateActorId(self) -> ObjectId:
        return ObjectId()

    def _appendActorInfo(self, actorInfo: dict):
        self._safe_update_one(
            {"_id": self._workflowId}, {"$push": {"actors": actorInfo}}
        )

    def _sanitize(self, value: Any) -> Any:
        if isinstance(value, Mapping):
            return {k: self._sanitize(v) for k, v in value.items()}
        if isinstance(value, (list, tuple, set)):
            return [self._sanitize(v) for v in value]
        if not any(isinstance(value, t) for t in bson._BUILT_IN_TYPES):
            return repr(value)
        return value



import pytest
from . import utils
from . import filter_combinations


@pytest.mark.parametrize("task_name", list(utils.SUCCESS))
@pytest.mark.parametrize("context", utils.CONTEXTS)
@pytest.mark.parametrize("pool_type", utils.POOLS)
@pytest.mark.parametrize("scaling", [True, False])
def test_callback(task_name, context, pool_type, scaling):
    with utils.pool_context(scaling, pool_type, context=context) as pool:
        with filter_combinations.filter_callback(pool_type, task_name, context):
            utils.assert_callback(pool, task_name)


@pytest.mark.parametrize("task_name", list(utils.FAILURE))
@pytest.mark.parametrize("context", utils.CONTEXTS)
@pytest.mark.parametrize("pool_type", utils.POOLS)
@pytest.mark.parametrize("scaling", [True, False])
def test_error_callback(task_name, context, pool_type, scaling):
    with utils.pool_context(scaling, pool_type, context=context) as pool:
        with filter_combinations.filter_error_callback(pool_type, task_name, context):
            utils.assert_error_callback(pool, task_name)



import pytest
from . import utils


def add(a, b):
    return a + b


def error(a, b):
    raise RuntimeError("intentional error")


FUNCTIONS = {"add": add, "error": error}


@pytest.mark.parametrize("scaling", [True, False])
@pytest.mark.parametrize("max_workers", [None, 1])
@pytest.mark.parametrize("pool_type", utils.POOLS)
@pytest.mark.parametrize("func", ["add", "error"])
def test_apply_async(scaling, max_workers, pool_type, func):
    callback_event = utils.Event()
    failed_msg = ""

    def result_callback(return_value):
        nonlocal failed_msg
        try:
            if return_value != 2:
                failed_msg = f"{return_value} != 2"
        finally:
            callback_event.set()

    def error_callback(exception):
        nonlocal failed_msg
        try:
            if not isinstance(exception, RuntimeError):
                failed_msg = f"{exception} is not a RuntimeError"
            elif str(exception) != "intentional error":
                failed_msg = f"'{exception}' != 'intentional error'"
        finally:
            callback_event.set()

    with utils.pool_context(scaling, pool_type, max_workers=max_workers) as pool:
        pool.apply_async(
            FUNCTIONS[func],
            args=(1, 1),
            callback=result_callback,
            error_callback=error_callback,
        )
        callback_event.wait(10)
        assert not failed_msg, failed_msg



from pypushflow.Workflow import Workflow
from pypushflow.Submodel import Submodel
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Submodel6(Submodel):
    """
    Submodel containing one python actor which has a long execution time
    """

    def __init__(self, parent, name, thread_counter):
        super().__init__(parent=parent, name=name, thread_counter=thread_counter)
        self.pythonActorAdd = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Submodel Actor Add",
            errorHandler=self,
            thread_counter=thread_counter,
        )
        self.getPort("In").connect(self.pythonActorAdd)
        self.pythonActorAdd.connect(self.getPort("Out"))


class Workflow6(Workflow):
    """
    Workflow containing one start actor,
    one submodel which has a long execution and one stop actor with short timeout.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(self, thread_counter=ctr)
        self.pythonActorAdd1 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Add Actor 1",
            thread_counter=ctr,
        )
        self.pythonActorAdd2 = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAdd.py",
            name="Python Add Actor 2",
            thread_counter=ctr,
        )
        self.submodel6 = Submodel6(self, name="Submodel 6", thread_counter=ctr)
        self.stopActor = StopActor(self, thread_counter=ctr)
        self.startActor.connect(self.pythonActorAdd1)
        self.pythonActorAdd1.connect(self.submodel6.getPort("In"))
        self.submodel6.getPort("Out").connect(self.pythonActorAdd2)
        self.pythonActorAdd2.connect(self.stopActor)


class TestWorkflow6(WorkflowTestCase):
    def test_workflow6(self):
        workflow6 = Workflow6("Test workflow 6")
        inData = {"value": 1}
        outData = workflow6.run(
            inData, timeout=5, scaling_workers=False, max_workers=-1
        )
        self.assertEqual(outData["value"], 4)



from pypushflow.AbstractActor import AbstractActor


class ErrorHandler(AbstractActor):
    def __init__(self, parent=None, name="Error handler", **kw):
        super().__init__(parent=parent, name=name, **kw)

    def trigger(self, inData):
        if self.parent is not None:
            self.parent.setStatus("error")
        super().trigger(inData=inData)



import sys
from multiprocessing import context


def get_context(context: str = None) -> context.BaseContext:
    return _NONDAEMONIC_CONTEXTS[context]()


class SpawnNonDaemonicProcess(context.SpawnProcess):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class SpawnNonDaemonicContext(context.SpawnContext):
    Process = SpawnNonDaemonicProcess


if sys.platform == "win32":
    _NONDAEMONIC_CONTEXTS = {
        None: SpawnNonDaemonicContext,
        "spawn": SpawnNonDaemonicContext,
    }
else:

    class ForkNonDaemonicProcess(context.ForkProcess):
        @property
        def daemon(self):
            return False

        @daemon.setter
        def daemon(self, value):
            pass

    class ForkServerNonDaemonicProcess(context.ForkServerProcess):
        @property
        def daemon(self):
            return False

        @daemon.setter
        def daemon(self, value):
            pass

    class ForkNonDaemonicContext(context.ForkContext):
        Process = ForkNonDaemonicProcess

    class ForkServerNonDaemonicContext(context.ForkServerContext):
        Process = ForkServerNonDaemonicProcess

    if sys.platform == "darwin":
        _NONDAEMONIC_CONTEXTS = {
            None: SpawnNonDaemonicContext,
            "spawn": SpawnNonDaemonicContext,
            "fork": ForkNonDaemonicContext,
            "forkserver": ForkServerNonDaemonicContext,
        }
    else:
        _NONDAEMONIC_CONTEXTS = {
            None: ForkNonDaemonicContext,
            "spawn": SpawnNonDaemonicContext,
            "fork": ForkNonDaemonicContext,
            "forkserver": ForkServerNonDaemonicContext,
        }



from pypushflow.Workflow import Workflow
from pypushflow.Submodel import Submodel
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Submodel4(Submodel):
    """
    Submodel containing one python actor which throws an exception.
    """

    def __init__(self, parent, name, thread_counter):
        super().__init__(parent=parent, name=name, thread_counter=thread_counter)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonErrorHandlerTest.py",
            name="Python Error Handler Test",
            errorHandler=self,
            thread_counter=thread_counter,
        )
        self.getPort("In").connect(self.pythonActor)
        self.pythonActor.connect(self.getPort("Out"))


class Workflow4(Workflow):
    """
    Workflow containing one start actor,
    one submodel which throws an exception and one stop actor.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(thread_counter=ctr)
        self.submodel4 = Submodel4(parent=self, name="Submodel 4", thread_counter=ctr)
        self.stopActor = StopActor(thread_counter=ctr)
        self.startActor.connect(self.submodel4.getPort("In"))
        self.submodel4.getPort("Out").connect(self.stopActor)
        self.connectOnError(self.stopActor)


class TestWorkflow4(WorkflowTestCase):
    def test_workflow4(self):
        workflow4 = Workflow4("Test workflow 4")
        inData = {"name": "Dog"}
        outData = workflow4.run(
            inData, timeout=5, scaling_workers=False, max_workers=-1
        )
        self.assertIsNotNone(outData)
        self.assertTrue("WorkflowException" in outData)



from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Workflow1(Workflow):
    """
    Workflow containing one start actor,
    one python actor and one stop actor.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorTest.py",
            name="Python Actor Test",
            thread_counter=ctr,
        )
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.pythonActor)
        self.pythonActor.connect(self.stopActor)


class TestWorkflow1(WorkflowTestCase):
    def test_Workflow1(self):
        testWorkflow1 = Workflow1("Test workflow 1")
        inData = {"name": "Jerry"}
        outData = testWorkflow1.run(
            inData, timeout=15, scaling_workers=False, max_workers=-1
        )
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Jerry!")



from time import sleep


def run(**kwargs):
    counter = kwargs.setdefault("counter", 0)
    print(f"SLEEP counter={counter} ...")
    sleep(kwargs.get("sleep_time", 0))
    print(f"DONE counter={counter}")
    kwargs["counter"] += 1
    return kwargs



from pypushflow.Workflow import Workflow
from pypushflow.Submodel import Submodel
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Submodel1(Submodel):
    """
    Submodel containing one python actor.
    """

    def __init__(self, parent, name, thread_counter):
        super().__init__(parent=parent, name=name, thread_counter=thread_counter)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorTest.py",
            name="Python Actor Test",
            thread_counter=thread_counter,
        )
        self.getPort("In").connect(self.pythonActor)
        self.pythonActor.connect(self.getPort("Out"))


class Workflow3(Workflow):
    """
    Workflow containing one start actor,
    one submodel and one stop actor.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.submodel1 = Submodel1(parent=self, name="Submodel 1", thread_counter=ctr)
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.submodel1.getPort("In"))
        self.submodel1.getPort("Out").connect(self.stopActor)


class TestWorkflow3(WorkflowTestCase):
    def test_workflow3(self):
        testWorkflow3 = Workflow3("Test workflow 3")
        inData = {"name": "Cat"}
        outData = testWorkflow3.run(
            inData, timeout=15, scaling_workers=False, max_workers=-1
        )
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Cat!")



"""Utilities to stop workflows on certain signals"""

import os
import signal
from weakref import WeakKeyDictionary
from typing import Optional, Sequence
from contextlib import contextmanager

if os.name == "nt":
    DEFAULT_STOP_SIGNALS = (signal.SIGINT,)
else:
    DEFAULT_STOP_SIGNALS = (signal.SIGTERM,)


class StopPypushflowWorkflow(Exception):
    pass


@contextmanager
def stop_on_signals_context(workflow, stop_signals: Optional[Sequence] = None):
    if workflow in _WORKFLOWS:
        yield
        return
    if stop_signals is None:
        stop_signals = DEFAULT_STOP_SIGNALS
    _WORKFLOWS[workflow] = stop_signals
    try:
        for signum in stop_signals:
            _stop_workflows_on_signal(signum)
        yield
    finally:
        _WORKFLOWS.pop(workflow)
        _reset_signals_handlers()


_WORKFLOWS = WeakKeyDictionary()
_OLD_HANDLERS = dict()


def _stop_workflows_on_signal(signum):
    if signum in _OLD_HANDLERS:
        return

    def stop_workflows_handler(signum, frame):
        for workflow, stop_signals in list(_WORKFLOWS.items()):
            if signum not in stop_signals:
                continue
            workflow.stop(f"stop workflow due to signal {signum}")

    _OLD_HANDLERS[signum] = _set_handler(signum, stop_workflows_handler)


def _reset_signals_handlers():
    keep = {signum for lst in list(_WORKFLOWS.values()) for signum in lst}
    remove = set(_OLD_HANDLERS) - keep
    for signum in remove:
        old_handler = _OLD_HANDLERS.pop(signum)
        _set_handler(signum, old_handler)


def _set_handler(signum, handler):
    try:
        old_handler = signal.signal(signum, handler)
    except (OSError, AttributeError, ValueError, RuntimeError):
        pass
    if old_handler is None:
        return signal.SIG_IGN
    return old_handler



from pypushflow.AbstractActor import AbstractActor


class StartActor(AbstractActor):
    def __init__(self, parent=None, name="Start actor", **kw):
        super().__init__(parent=parent, name=name, **kw)



from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Workflow2(Workflow):
    """
    Workflow with error handling, containing one start actor,
    one python actor and one stop actor.

    The python actor throws an exception.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonErrorHandlerTest.py",
            name="Python Error Handler Test",
            errorHandler=self,
            thread_counter=ctr,
        )
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.pythonActor)
        self.pythonActor.connect(self.stopActor)
        self.connectOnError(self.stopActor)


class TestWorkflow2(WorkflowTestCase):
    def test_Workflow2(self):
        testWorkflow2 = Workflow2("Test workflow 2")
        inData = {"name": "Tom"}
        outData = testWorkflow2.run(inData, timeout=5, scaling_workers=False)
        self.assertIsNotNone(outData)
        self.assertTrue("WorkflowException" in outData)



import pprint
from pypushflow.AbstractActor import AbstractActor


class JoinActor(AbstractActor):
    def __init__(self, parent=None, name="Join actor", **kw):
        super().__init__(parent=parent, name=name, **kw)
        self.numberOfThreads = 0
        self.listInData = []

    def increaseNumberOfThreads(self):
        self.numberOfThreads += 1

    def trigger(self, inData):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        self.setStarted()
        self.setFinished()
        self.listInData.append(inData)
        if len(self.listInData) == self.numberOfThreads:
            newInData = {}
            for data in self.listInData:
                newInData.update(data)
            for actor in self.listDownStreamActor:
                actor.trigger(newInData)



import time
from numbers import Number
from typing import Callable, Optional
from .base import BasePool


class ScalingPool(BasePool):
    def __init__(
        self,
        wait_on_exit: bool = True,
        wait_on_exit_timeout: Optional[Number] = None,
        interrupt_on_exit: bool = False,
        pool_type: Optional[str] = None,
        **pool_options,
    ):
        if pool_type == "scaling":
            raise ValueError("cannot nest scaling pools")
        self._running_pools = list()
        self._finished_pools = list()
        self._pool_type = pool_type
        pool_options["max_workers"] = 1
        self._pool_options = pool_options
        super().__init__(
            wait_on_exit=wait_on_exit,
            wait_on_exit_timeout=wait_on_exit_timeout,
            interrupt_on_exit=interrupt_on_exit,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)

        for pool in self._running_pools:
            pool.__exit__(exc_type, exc_val, exc_tb)
        self._running_pools = list()

        for pool in self._finished_pools:
            pool.__exit__(exc_type, exc_val, exc_tb)
        self._finished_pools = list()

    def close(self):
        for pool in self._finished_pools:
            pool.close()
        for pool in self._running_pools:
            pool.close()

    def join(self, timeout: Optional[Number] = None):
        if timeout is not None:
            t0 = time.time()

        for pool in self._running_pools:
            if not pool.join(timeout=timeout):
                return False
            if timeout is not None:
                timeout -= time.time() - t0
        self._running_pools = list()

        for pool in self._finished_pools:
            if not pool.join(timeout=timeout):
                return False
            if timeout is not None:
                timeout -= time.time() - t0
        self._finished_pools = list()

    def interrupt(self) -> None:
        for pool in list(self._running_pools):
            pool.interrupt()
        for pool in list(self._finished_pools):
            pool.interrupt()

    def _new_pool(self):
        from .factory import get_pool

        pool = get_pool(self._pool_type)(**self._pool_options)
        self._running_pools.append(pool)
        return pool

    def _release_pool(self, pool):
        try:
            idx = self._running_pools.index(pool)
        except ValueError:
            return
        pool = self._running_pools.pop(idx)
        self._finished_pools.append(pool)

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ):
        pool = self._new_pool()

        if callback is None:

            def _callback(return_value):
                self._release_pool(pool)

        else:

            def _callback(return_value):
                try:
                    return callback(return_value)
                finally:
                    self._release_pool(pool)

        if callback is None:

            def _error_callback(exception):
                self._release_pool(pool)

        else:

            def _error_callback(exception):
                try:
                    return error_callback(exception)
                finally:
                    self._release_pool(pool)

        future = pool.apply_async(
            fn,
            args=args,
            kwargs=kwargs,
            callback=_callback,
            error_callback=_error_callback,
        )
        pool.close()
        return future



from numbers import Number
from gevent.pool import Group
from gevent import Greenlet
from typing import Callable, Optional
from . import base
from .interrupt import gevent as terminate


class GreenletPool(base.BasePool):
    """Pool of greenlets."""

    def __init__(self, **kw) -> None:
        self._closed = False
        self._pool = Group()
        super().__init__(**kw)

    def shutdown(
        self,
        block: bool = False,
        timeout: Optional[Number] = None,
        interrupt: bool = False,
    ) -> None:
        super().shutdown(block=block, timeout=timeout, interrupt=interrupt)
        self._pool.kill(block=block, timeout=timeout)

    def close(self):
        self._closed = True

    def join(self, timeout: Optional[Number] = None) -> bool:
        return self._pool.join(timeout=timeout, raise_error=False)

    def interrupt(self) -> None:
        for glts in list(self._pool.greenlets):
            terminate.interrupt_task(glts)

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ) -> Greenlet:
        if self._closed:
            raise RuntimeError("the pool is closed")
        if self._pool is None:
            raise RuntimeError("enter the pool context first")
        if kwargs is None:
            kwargs = dict()

        def cb(glt):
            try:
                result = glt.get()
            except Exception as e:
                if error_callback is not None:
                    error_callback(e)
            else:
                if callback is not None:
                    callback(result)

        glt = Greenlet(terminate.task_main, fn, *args, **kwargs)
        glt.link(cb)
        self._pool.start(glt)
        return glt



import sys
import logging
from numbers import Number
from typing import Callable, Optional
from concurrent.futures import ProcessPoolExecutor, Future
from multiprocessing import get_context
from multiprocessing import set_start_method

from . import base
from .interrupt import process as interrupt

logger = logging.getLogger(__name__)


class ProcessPool(base.BasePool):
    """Pool of daemonic processes (they cannot have sub-processes)."""

    def __init__(
        self, context: str = None, max_workers: Optional[int] = None, **kw
    ) -> None:
        kwargs = dict()

        if sys.version_info >= (3, 7):
            if isinstance(context, str) or None:
                context = get_context(context)
            logger.info(f"pypushflow process pool context: '{type(context).__name__}'")
            kwargs["mp_context"] = context
        else:
            logger.info(f"pypushflow process pool context: '{context}'")
            assert isinstance(context, str) or context is None
            set_start_method(context, force=True)
        if max_workers is not None:
            kwargs["max_workers"] = max_workers
        self._pool = _initialize_pool(kwargs)
        self._closed = False
        super().__init__(**kw)

    def __enter__(self):
        self._pool.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self._pool.__exit__(exc_type, exc_val, exc_tb)

    def shutdown(
        self,
        block: bool = False,
        timeout: Optional[Number] = None,
        interrupt: bool = False,
    ) -> None:
        super().shutdown(block=block, timeout=timeout, interrupt=interrupt)
        self._pool.shutdown(wait=block)

    def close(self):
        self._closed = True

    def join(self, timeout: Optional[Number] = None) -> bool:
        return False

    def interrupt(self) -> None:
        for pid in list(self._pool._processes):
            interrupt.interrupt_task(pid)

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ) -> Future:
        if self._closed:
            raise RuntimeError("the pool is closed")

        def cb(future):
            try:
                result = future.result()
            except Exception as e:
                if error_callback is not None:
                    error_callback(e)
            else:
                if callback is not None:
                    callback(result)

        if kwargs is None:
            kwargs = dict()
        future = self._pool.submit(interrupt.task_main, fn, *args, **kwargs)
        future.add_done_callback(cb)
        return future


def _initialize_pool(kwargs) -> ProcessPoolExecutor:
    if sys.version_info >= (3, 7):
        return ProcessPoolExecutor(initializer=interrupt.worker_initializer, **kwargs)

    pool = ProcessPoolExecutor(**kwargs)
    pool.submit(_worker_initializer, 0).result()
    pool.map(_worker_initializer, [1] * len(pool._processes))
    return pool


def _worker_initializer(seconds):
    from time import sleep

    interrupt.worker_initializer()
    sleep(seconds)



import sys
from concurrent.futures import ProcessPoolExecutor as CfPool

from multiprocessing.pool import Pool as MpPool
from multiprocessing import Value as MpValue
from multiprocessing import get_context as get_mp_context
from multiprocessing import current_process as current_mp_process
from multiprocessing import get_all_start_methods
from multiprocessing import set_start_method

from billiard.pool import Pool as BPool
from billiard import get_context as get_billiard_context
from billiard import Value as BValue
from billiard import current_process as current_b_process

from pypushflow.concurrent.non_daemonic import get_context as get_nd_context


def current_process_is_daemonic(result=None, billiard=False):
    if billiard:
        func = current_b_process
    else:
        func = current_mp_process
    if result is None:
        return func().daemon
    else:
        result.value = func().daemon


def exec_mppool(context=None):
    with MpPool(context=get_mp_context(context)) as pool:
        return pool.apply(current_process_is_daemonic)


def exec_nd_mppool(context=None):
    with MpPool(context=get_nd_context(context)) as pool:
        return pool.apply(current_process_is_daemonic)


def exec_bpool(context=None):
    with BPool(context=get_billiard_context(context)) as pool:
        return pool.apply(current_process_is_daemonic, kwds={"billiard": True})


def exec_cfpool(context=None):
    if sys.version_info >= (3, 7):
        with CfPool(mp_context=get_mp_context(context)) as pool:
            return pool.submit(current_process_is_daemonic).result()
    else:
        set_start_method(context, force=True)
        with CfPool() as pool:
            return pool.submit(current_process_is_daemonic).result()


def exec_nd_cfpool(context=None):
    if sys.version_info >= (3, 7):
        with CfPool(mp_context=get_nd_context(context)) as pool:
            return pool.submit(current_process_is_daemonic).result()


def exec_mpprocess(context=None):
    result = MpValue("i", 0)
    p = get_mp_context(context).Process(
        target=current_process_is_daemonic, args=(result,)
    )
    p.start()
    p.join()
    return bool(result.value)


def exec_nd_mpprocess(context=None):
    result = MpValue("i", 0)
    p = get_nd_context(context).Process(
        target=current_process_is_daemonic, args=(result,)
    )
    p.start()
    p.join()
    return bool(result.value)


def exec_bprocess(context=None):
    result = BValue("i", 0)
    p = get_billiard_context(context).Process(
        target=current_process_is_daemonic, args=(result,), kwargs={"billiard": True}
    )
    p.start()
    p.join()
    return bool(result.value)


if __name__ == "__main__":
    daemonic = {True: list(), False: list(), None: list()}
    for context in get_all_start_methods():
        for func in (
            exec_mppool,
            exec_nd_mppool,
            exec_bpool,
            exec_cfpool,
            exec_nd_cfpool,
            exec_mpprocess,
            exec_nd_mpprocess,
            exec_bprocess,
        ):
            daemonic[func(context)].append(f" {func.__name__}('{context}')")

    print("Daemonic:")
    for name in daemonic[True]:
        print(name)
    print("Non-Daemonic:")
    for name in daemonic[False]:
        print(name)
    print("Not supported:")
    for name in daemonic.get(None):
        print(name)



from concurrent.futures import ThreadPoolExecutor, Future
from numbers import Number
from typing import Callable, Optional
from . import base


class ThreadPool(base.BasePool):
    def __init__(self, max_workers: Optional[int] = None, **kw) -> None:
        kwargs = dict()
        if max_workers is not None:
            kwargs["max_workers"] = max_workers
        self._pool = ThreadPoolExecutor(**kwargs)
        self._closed = False
        super().__init__(**kw)

    def __enter__(self):
        self._pool.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self._pool.__exit__(exc_type, exc_val, exc_tb)

    def close(self):
        self._closed = True

    def join(self, timeout: Optional[Number] = None) -> bool:
        return False

    def interrupt(self) -> None:
        raise RuntimeError("Cannot stop tasks of a thread pool")

    def apply_async(
        self,
        fn: Callable,
        callback: Optional[Callable] = None,
        error_callback: Optional[Callable] = None,
        args=tuple(),
        kwargs=None,
    ) -> Future:
        if self._closed:
            raise RuntimeError("the pool is closed")

        def cb(future):
            try:
                result = future.result()
            except Exception as e:
                if error_callback is not None:
                    error_callback(e)
            else:
                if callback is not None:
                    callback(result)

        if kwargs is None:
            kwargs = dict()
        future = self._pool.submit(fn, *args, **kwargs)
        future.add_done_callback(cb)
        return future



"""Utilities to terminate process pool tasks"""

import os
import signal
from contextlib import contextmanager
from . import StopPypushflowTask

_TERMINATE_SIGNAL = signal.SIGINT


def task_main(fn, *args, **kwargs):
    with _task_context():
        return fn(*args, **kwargs)


def interrupt_task(pid) -> None:
    # TODO: on windows this kills the subprocess
    #       despite that fact that we handle the
    #       signal in the subprocess
    try:
        os.kill(pid, _TERMINATE_SIGNAL)
    except ProcessLookupError:
        pass  # process already ended


def worker_initializer():
    _set_terminate_handler(signal.SIG_IGN)


@contextmanager
def _task_context():
    def stop_task_handler(signum, frame):
        nonlocal old_handler
        if old_handler is None:
            old_handler = signal.SIG_IGN
        signal.signal(signum, old_handler)
        raise KeyboardInterrupt(f"stop task due to signal {signum}")

    old_handler = _set_terminate_handler(stop_task_handler)

    try:
        yield
    except KeyboardInterrupt as e:
        raise StopPypushflowTask(str(e)) from None
    finally:
        _set_terminate_handler(old_handler)


def _set_terminate_handler(handler):
    try:
        return signal.signal(_TERMINATE_SIGNAL, handler)
    except (OSError, AttributeError, ValueError, RuntimeError):
        pass



from . import multiprocessing
from .non_daemonic import get_context


class NProcessPool(multiprocessing.MProcessPool):
    """Pool of non-daemonic processes (they can have sub-processes)."""

    def __init__(self, context: str = None, **kw) -> None:
        context = get_context(context)
        super().__init__(context=context, **kw)



"""Utilities to terminate gevent pool tasks"""

from gevent import GreenletExit
from . import StopPypushflowTask


def task_main(fn, *args, **kwargs):
    try:
        return fn(*args, **kwargs)
    except GreenletExit:
        raise StopPypushflowTask("stop task due to interrupt") from None


def interrupt_task(glt) -> None:
    glt.kill()



import pprint
from typing import Optional, Sequence
from contextlib import contextmanager
from .persistence import init_db_client
from .logutils import PyPushflowLoggedObject
from .concurrent import get_pool
from . import stop_workflows


class Workflow(PyPushflowLoggedObject):
    def __init__(
        self,
        name,
        stop_on_signals: bool = False,
        forced_interruption: bool = False,
        stop_signals: Optional[Sequence] = None,
        db_options: Optional[dict] = None,
    ):
        """
        :param name:
        :param level:
        :param stop_on_signals:
        :param forced_interruption:
        :param stop_signals:
        """
        super().__init__(log_metadata={"workflow": name})
        self.logger.info("\n\nInstantiate new workflow '%s'\n", name)
        self.name = name
        self.listOnErrorActor = []
        if db_options is None:
            db_options = dict()
        self.db_client = init_db_client(**db_options)
        self.listActorRef = []
        self._execution_pool = None
        self._stop_exception = None
        self._stop_on_signals = stop_on_signals
        self._forced_interruption = forced_interruption
        self._stop_signals = stop_signals

    def connectOnError(self, actor):
        self.logger.debug("connect to error handler '%s'", actor.name)
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, inData):
        self.logger.info(
            "triggered due to error with inData =\n %s", pprint.pformat(inData)
        )
        for onErrorActor in self.listOnErrorActor:
            onErrorActor.trigger(inData)

    def getActorPath(self):
        return "/" + self.name

    def addActorRef(self, actorRef):
        self.logger.debug("add reference to actor '%s'", actorRef.name)
        self.listActorRef.append(actorRef)

    def getListActorRef(self):
        return self.listActorRef

    def setStatus(self, status):
        self.db_client.setWorkflowStatus(status)

    def endWorkflow(self, status):
        self.db_client.endWorkflow(status)

    @contextmanager
    def _run_context(self, **pool_options):
        with self._pool_context(**pool_options):
            with self._db_client_context():
                yield

    @contextmanager
    def _pool_context(
        self,
        max_workers: Optional[int] = None,
        scaling_workers: bool = True,
        pool_type: Optional[str] = None,
        **pool_options
    ):
        if self._execution_pool is not None:
            # A pool already exists
            yield
            return
        if max_workers is not None and max_workers <= 0:
            max_workers = sum(actor.pool_resources for actor in self.listActorRef)
        pool_options["max_workers"] = max_workers
        pool_options["pool_type"] = pool_type
        if scaling_workers:
            pool_type = "scaling"
        try:
            with self._stop_on_signals_context():
                with get_pool(pool_type)(**pool_options) as pool:
                    self.logger.info(
                        "Execution pool %s (%s)", type(pool).__name__, pool_options
                    )
                    self._execution_pool = pool
                    yield
        except BaseException as e:
            # Make sure PythonActor callbacks do not trigger other tasks
            self.logger.warning("Interrupt workflow execution: %s", str(e))
            self._stop_exception = e
            raise
        finally:
            if self._execution_pool is not None:
                self._execution_pool.join()
                self._execution_pool = None
            self._stop_exception = None

    @contextmanager
    def _db_client_context(self):
        self.db_client.connect()
        try:
            self.db_client.startWorkflow(self.name)
            try:
                yield
            finally:
                self.db_client.ensureEndWorkflow()
        finally:
            self.db_client.disconnect()

    @contextmanager
    def _stop_on_signals_context(self):
        if self._stop_on_signals:
            with stop_workflows.stop_on_signals_context(
                self, stop_signals=self._stop_signals
            ):
                yield
        else:
            yield

    def run(self, inData, timeout=None, **pool_options):
        self.logger.info("\n\nRun workflow '%s'\n", self.name)
        self.stopActor.reset()
        with self._run_context(**pool_options):
            self.startActor.trigger(inData)
            self.stopActor.join(timeout=timeout)
            return self.stopActor.outData

    @property
    def pool(self):
        return self._execution_pool

    @property
    def stop_exception(self) -> Optional[BaseException]:
        if self._stop_exception:
            return self._stop_exception

    def stop(
        self,
        reason: str = "interrupt workflow",
        forced_interruption: Optional[bool] = None,
    ):
        self.logger.debug("stop workflow: %s", reason)
        if forced_interruption is None:
            forced_interruption = self._forced_interruption
        self._stop_exception = stop_workflows.StopPypushflowWorkflow(reason)
        if forced_interruption and self.pool is not None:
            self.pool.close()
            self.pool.interrupt()



import time


def run(all_arguments=None):
    if all_arguments is None:
        raise RuntimeError("Missing argument 'all_arguments'!")
    time.sleep(1)
    value = all_arguments["value"]
    value = value + 1
    all_arguments["value"] = value
    return {"all_arguments": all_arguments}



import traceback


def serialize_exception(exception: BaseException) -> dict:
    return {
        "errorMessage": str(exception),
        "traceBack": traceback.format_exception(
            type(exception), exception, exception.__traceback__
        ),
    }



from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.RouterActor import RouterActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.tests.workflowTestCase import WorkflowTestCase


class Workflow10(Workflow):
    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(self, thread_counter=ctr)
        self.pythonActorAddWithoutSleep = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorAddWithoutSleep.py",
            name="Add without sleep",
            thread_counter=ctr,
        )
        self.pythonActorCheck = PythonActor(
            parent=self,
            script="pypushflow.tests.tasks.pythonActorCheck.py",
            name="Check",
            thread_counter=ctr,
        )
        self.check = RouterActor(
            parent=self,
            name="Check",
            itemName="doContinue",
            listPort=["true", "false"],
            thread_counter=ctr,
        )
        self.stopActor = StopActor(self, thread_counter=ctr)
        self.startActor.connect(self.pythonActorAddWithoutSleep)
        self.pythonActorAddWithoutSleep.connect(self.pythonActorCheck)
        self.pythonActorCheck.connect(self.check)
        self.check.connect(self.pythonActorAddWithoutSleep, expectedValue="true")
        self.check.connect(self.stopActor, expectedValue="false")


class TestWorkflow10(WorkflowTestCase):
    def test_workflow10(self):
        limit = 10
        workflow10 = Workflow10(f"Test workflow {limit}")
        inData = {"value": 1, "limit": limit}
        outData = workflow10.run(
            inData, timeout=1200, scaling_workers=False, max_workers=-1
        )
        self.assertEqual(outData["value"], limit)
